import json
import subprocess
import re
import base64

_RETURN_MSG_PATTERN = re.compile(r"##(.*)##(.*)##$")


class RunResult:
    def __init__(self, status: bool, js_type=None, output=None, error=None):
        self.status = status
        self.js_type = js_type
        self.output = output
        self.error = error


def call_js(js_script: str, call_func_name: str, call_func_params=None):
    """
        Call js_script with call_func_name and call_func_params
    """
    expression = make_call_expression(call_func_name, call_func_params)
    return run_js(js_script, expression)


def run_js(js_script: str, expression: str):
    """
        在 js_script 脚本基础上运行 js 表达式，并获得表达式返回值
    """
    result = run_js_with_result(js_script, expression)
    if result.status:
        return result.output
    else:
        msg = "run js error: " + result.error
        raise RuntimeError(msg)


def run_js_with_result(js_script: str, expression: str):
    if js_script is None or expression is None:
        raise ValueError('js_script and expression are required')
    js_script = base64.b64encode(js_script.encode('utf-8')).decode('utf-8')
    js_code = f'''
        (function() {{
            eval(Buffer.from("{js_script}", "base64").toString("UTF-8"));
            !(function(){{
                let _handle_result = function(result) {{
                    let result_type = typeof result;
                    if (result_type !== 'string') {{
                        result = JSON.stringify(result);
                    }}
                    let result_base64 = Buffer.from(result, "UTF-8").toString("base64");
                    let result_type_base64 = Buffer.from(result_type, "UTF-8").toString("base64");
                    process.stdout.write(`##${{result_type_base64}}##${{result_base64}}##`);
                }}
                let expression_result = {expression};
                if (expression_result && typeof expression_result.then === 'function') {{
                    expression_result.then(obj => {{_handle_result(obj)}});
                }} else {{
                    _handle_result(expression_result);
                }}
            }})()
        }})()
    '''
    p = subprocess.Popen(['node'],
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate(js_code.encode('ascii'))
    ok = p.wait() == 0
    result = RunResult(ok)
    if ok:
        g = _RETURN_MSG_PATTERN.search(out.decode('ascii'))
        result.js_type = base64.b64decode(g.group(1)).decode('utf-8')
        result.output = base64.b64decode(g.group(2)).decode('utf-8')
    else:
        result.error = err.decode()
    return result


def make_call_expression(call_func_name, call_func_params):
    """
        @param call_func_name: Call function name
        @param call_func_params: Call function parameters
    """
    if not isinstance(call_func_name, str):
        raise RuntimeError(f"call_func_name must be a string")
    _call_func_params_is_list_or_tuple = False
    if not isinstance(call_func_params, (str, int, dict)):
        if isinstance(call_func_params, (list, tuple)):
            _call_func_params_is_list_or_tuple = True
            for param in call_func_params:
                if not isinstance(param, (str, int, dict)) and param is not None:
                    raise RuntimeError(f"call_func_params every param must be (str, int, dict) or None")
        elif call_func_params is not None:
            raise RuntimeError(f"call_func_params must be (str, int, dict) or None")
    if _call_func_params_is_list_or_tuple:
        params = [json.dumps(param, ensure_ascii=False) for param in call_func_params]
        expression = '{}({})'.format(call_func_name, ','.join(params))
    else:
        expression = '{}({})'.format(call_func_name, json.dumps(call_func_params, ensure_ascii=False))
    return expression



import unittest
import pyrunjs


class TestRunJs(unittest.TestCase):

    def test_hello_world(self):
        js_script = """function HelloWorld() { return 'Hello World!'; } """
        self.assertEqual(pyrunjs.run_js(js_script, "HelloWorld()"), "Hello World!")

    def test_add(self):
        self.assertEqual(pyrunjs.run_js('', "1 + 2"), "3")

    def test_run_error(self):
        with self.assertRaises(RuntimeError):
            pyrunjs.run_js('', "let let a = 1;")


if __name__ == '__main__':
    unittest.main()



import unittest

import pyrunjs


class TestRunJsPromise(unittest.TestCase):

    def test_resolve_promise(self):
        js_script = """
            function add(a, b) {
                return new Promise((resolve, reject) => {
                    let sum = a + b;
                    resolve(sum);
                });
            }
        """
        self.assertEqual("3", pyrunjs.run_js(js_script, "add(1, 2)"))
        self.assertEqual("3", pyrunjs.call_js(js_script, "add", (1, 2)))

    def test_reject_promise(self):
        js_script = """
            function add(a, b) {
                return new Promise((resolve, reject) => {
                    reject('this is reject error');
                });
            }
        """
        with self.assertRaises(RuntimeError):
            pyrunjs.run_js(js_script, "add(1, 2)")
        with self.assertRaises(RuntimeError):
            pyrunjs.call_js(js_script, "add", (1, 2))

    def test_promise_then(self):
        js_script = """
            function add(a, b) {
                return new Promise((resolve, reject) => {
                    let sum = a + b;
                    resolve(sum);
                }).then(sum => {
                    return sum * sum;
                });
            }
        """
        self.assertEqual("9", pyrunjs.run_js(js_script, "add(1, 2)"))


if __name__ == '__main__':
    unittest.main()



import json
import unittest
import pyrunjs


class TestCallJs(unittest.TestCase):

    def test_hello_world(self):
        js_script = """function HelloWorld() { return 'Hello World!'; } """
        self.assertEqual("Hello World!", pyrunjs.call_js(js_script, "HelloWorld"))

    def test_string_param(self):
        js_script = """function add(a) { return a; } """
        self.assertEqual("my msg.", pyrunjs.call_js(js_script, "add", "my msg."))

    def test_int_param(self):
        js_script = """function add(a) { return 1 + a; } """
        self.assertEqual("10", pyrunjs.call_js(js_script, "add", 9))

    def test_list_and_tuple_params(self):
        js_script = """function add(a, b, c) { return a + b - c; } """
        list_params = [1, 2, 3]
        self.assertEqual("0", pyrunjs.call_js(js_script, "add", list_params))
        tuple_params = (1, 2, 3)
        self.assertEqual("0", pyrunjs.call_js(js_script, "add", tuple_params))

    def test_dict_param(self):
        js_script = """function get_msg(a) { return a.msg; } """
        dict_params = {
            "key1": "value1",
            "key2": "value2",
            "msg": "hi"
        }
        self.assertEqual("hi", pyrunjs.call_js(js_script, "get_msg", dict_params), "msg should be hi")

    def test_tuple_param_with_dict(self):
        js_script = """function makeJsObj(a, b, c, d) { 
            a.key1 = a.key1 + b; 
            a.key2 = a.key2 + c;
            a.msg = d;
            return a; 
         } """
        dict_params = {
            "key1": "value1",
            "key2": 1,
            "msg": "hi"
        }
        params = (dict_params, 23, 1, "hello")
        result = pyrunjs.call_js(js_script, "makeJsObj", params)
        expected_dict = {
            "key1": "value123",
            "key2": 2,
            "msg": "hello"
        }
        self.assertEqual(json.loads(result), expected_dict)


if __name__ == '__main__':
    unittest.main()
